# Copyright 2023-2024 Luminary Cloud, Inc. All Rights Reserved.

import logging
from typing import Any, Iterable, Optional, Tuple, cast

import grpc
import re

from .._auth import Auth0Client
from .._proto.mesh.mesh_pb2_grpc import MeshServiceStub
from .._proto.project.project_pb2_grpc import ProjectServiceStub
from .._proto.simulation.simulation_pb2_grpc import SimulationServiceStub
from .auth0_metadata_plugin import Auth0MetadataPlugin
from .config import LC_DOMAIN
from .logging_interceptor import LoggingInterceptor
from .retry_interceptor import RetryInterceptor
from .tracing import add_instrumentation

logger = logging.getLogger(__name__)


class Client(
    ProjectServiceStub,
    MeshServiceStub,
    SimulationServiceStub,
):
    """
    Creates a Luminary API client.

    The returned client automatically obtains access tokens for the Luminary API and
    sends them with each RPC call. See auth/auth.py for details.

    Supports "with" syntax.

    Parameters
    ----------
    target : str
        The URL of the API server.
    localhost : bool
        True if the API server is running locally.
    grpc_channel_options : Optional[Iterable[tuple[str, str]]]
        A list of gRPC channel args. The full list is available here:
        https://github.com/grpc/grpc/blob/v1.46.x/include/grpc/impl/codegen/grpc_types.h
    **kwargs : dict, optional
        Additional arguments are passed to Auth0Client. See auth/auth.py.
    """

    def __init__(
        self,
        target: str = LC_DOMAIN,
        localhost: bool = False,
        grpc_channel_options: Optional[Iterable[Tuple[str, str]]] = None,
        channel_credentials: Optional[grpc.ChannelCredentials] = None,
        **kwargs: Any,
    ):
        self._target = target
        self._apiserver_domain = target.split(":", maxsplit=1)[0]
        self._auth0_client = Auth0Client(**kwargs)
        self._channel = self._create_channel(localhost, grpc_channel_options, channel_credentials)
        ProjectServiceStub.__init__(self, self._channel)
        MeshServiceStub.__init__(self, self._channel)
        SimulationServiceStub.__init__(self, self._channel)

    @property
    def channel(self) -> grpc.ChannelCredentials:
        return self._channel

    @property
    def apiserver_domain(self) -> str:
        return self._apiserver_domain

    @property
    def primary_domain(self) -> Optional[str]:
        return _get_primary_domain_for_apiserver_domain(self._apiserver_domain)

    @property
    def internal(self) -> bool:
        return _is_internal_domain_for_lc_apiserver(self._apiserver_domain)

    def get_token(self) -> str:
        return self._auth0_client.fetch_access_token()

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, *exc: Any) -> None:
        self._channel.close()

    def _create_channel(
        self,
        localhost: bool = False,
        grpc_channel_options: Optional[Iterable[Tuple[str, str]]] = None,
        channel_credentials: Optional[grpc.ChannelCredentials] = None,
    ) -> grpc.ChannelCredentials:
        if channel_credentials is None:
            if localhost:
                logger.debug("Using local channel credentials.")
                channel_credentials = grpc.local_channel_credentials()
            else:
                logger.debug("Using SSL channel credentials.")
                channel_credentials = grpc.ssl_channel_credentials()
        metadata_plugin = Auth0MetadataPlugin(self._auth0_client)
        auth0_credentials = grpc.metadata_call_credentials(metadata_plugin)
        options = grpc_channel_options and list(grpc_channel_options)
        channel = grpc.secure_channel(
            self._target,
            grpc.composite_channel_credentials(
                cast(grpc.ChannelCredentials, channel_credentials),
                auth0_credentials,
            ),
            options=options,
        )
        intercepted_channel = grpc.intercept_channel(
            channel,
            LoggingInterceptor(),
            RetryInterceptor(),
        )
        return add_instrumentation(
            intercepted_channel, self._target, self.primary_domain, self._auth0_client
        )


def _is_internal_domain_for_lc_apiserver(domain_name: str) -> bool:
    """Returns true iff the domain is an internal (non-prod) apiserver domain."""
    return re.match(r"apis[\.-].+\.luminarycloud\.com", domain_name) is not None


def _get_primary_domain_for_apiserver_domain(apiserver_domain: str) -> Optional[str]:
    """
    Get the frontend (primary) domain given an apiserver domain
    apis.luminarycloud.com -> app.luminarycloud.com
    apis-foo.int.luminarycloud.com -> foo.int.luminarycloud.com
    """
    if apiserver_domain == "apis.luminarycloud.com":  # prod
        return "app.luminarycloud.com"
    if apiserver_domain == "apis-itar.luminarycloud.com":  # itar-prod
        return "app-itar.luminarycloud.com"
    elif _is_internal_domain_for_lc_apiserver(apiserver_domain):
        return re.sub(r"^apis[-\.]{1}", "", apiserver_domain)
    return None
